@@ -36,8 +36,10 @@ import struct
 from fcntl import ioctl
 import termios
 import imp
+from livestreamer import Livestreamer
+from multiprocessing.pool import ThreadPool as Pool
+from multiprocessing import Manager
 
-rc = imp.new_module('rc')
 
 ID_FIELD_WIDTH   = 6
 NAME_FIELD_WIDTH = 22
@@ -46,11 +48,20 @@ VIEWS_FIELD_WIDTH = 7
 PLAYING_FIELD_OFFSET = ID_FIELD_WIDTH + NAME_FIELD_WIDTH + RES_FIELD_WIDTH + VIEWS_FIELD_WIDTH + 6
 
 PROG_STRING    = 'livestreamer-curses'
-VERSION_STRING = '1.0.1'
+VERSION_STRING = '1.1.0'
 TITLE_STRING   = '{0} v{1}'.format(PROG_STRING, VERSION_STRING)
 
 DEFAULT_RESOLUTION_HARD = '480p'
 
+INDICATORS = [
+        '  x  ', # offline
+        ' >>> ', # streaming
+        '  ?  ', # unknown
+        '  !  '  # error
+]
+
+CHECK_ONLINE_THREADS = 15
+
 class QueueFull(Exception): pass
 class QueueDuplicate(Exception): pass
 
@@ -147,7 +158,7 @@ class StreamPlayer(object):
 
 class StreamList(object):
 
-    def __init__(self, filename):
+    def __init__(self, filename, rc_module):
         """ Init and try to load a stream list, nothing about curses yet """
 
         # Open the storage (create it if necessary
@@ -160,20 +171,23 @@ class StreamList(object):
             for s in self.streams:
                 # Max id, needed when adding a new stream
                 self.max_id = max(self.max_id, s['id'])
+                s['online'] = 2
         else:
             self.streams = []
         self.filtered_streams = list(self.streams)
         self.filter = ''
+        self.show_offline_streams = False
+        self.rc_module = rc_module
 
-        if 'LIVESTREAMER_COMMANDS' in dir(rc):
-            self.cmd_list = map(shlex.split, rc.LIVESTREAMER_COMMANDS)
+        if 'LIVESTREAMER_COMMANDS' in dir(self.rc_module):
+            self.cmd_list = map(shlex.split, self.rc_module.LIVESTREAMER_COMMANDS)
         else:
             self.cmd_list = [['livestreamer']]
         self.cmd_index = 0
         self.cmd = self.cmd_list[self.cmd_index]
 
-        if 'DEFAULT_RESOLUTION' in dir(rc):
-            self.default_res = rc.DEFAULT_RESOLUTION
+        if 'DEFAULT_RESOLUTION' in dir(self.rc_module):
+            self.default_res = self.rc_module.DEFAULT_RESOLUTION
         else:
             self.default_res = DEFAULT_RESOLUTION_HARD
 
@@ -184,6 +198,8 @@ class StreamList(object):
         self.no_stream_shown = self.no_streams
         self.q = ProcessList(StreamPlayer().play)
 
+        self.livestreamer = Livestreamer()
+
     def __del__(self):
         """ Stop playing streams and sync storage """
         self.q.terminate()
@@ -204,6 +220,7 @@ class StreamList(object):
         curses.curs_set(0)
 
         self.s = s
+        self.s.keypad(1)
 
         self.get_screen_size()
 
@@ -215,13 +232,16 @@ class StreamList(object):
         self.current_pad = 'streams'
 
         self.set_title(TITLE_STRING)
-        self.set_footer('Ready')
 
         self.got_g = False
-        self.status = ''
 
         signal.signal(28, self.resize)
 
+        if 'CHECK_ONLINE_ON_START' in dir(self.rc_module) and self.rc_module.CHECK_ONLINE_ON_START:
+            self.check_online_streams()
+
+        self.set_status('Ready')
+
     def getheightwidth(self):
         """ getwidth() -> (int, int)
 
@@ -275,15 +295,14 @@ class StreamList(object):
                 if fd != sys.stdin:
                     # Set the new status line only if non-empty
                     msg = fd.readline()
-                    if len(msg) > 0:
-                        self.status = msg[:-1]
-                        self.redraw_status()
+                    if msg:
+                        self.set_status(msg[:-1])
                 else:
                     # Main event loop
                     c = self.pads[self.current_pad].getch()
-                    if c == curses.KEY_UP or c == ord('k') or c == 65:
+                    if c == curses.KEY_UP or c == ord('k'):
                         self.move(-1)
-                    elif c == curses.KEY_DOWN or c == ord('j') or c == 66:
+                    elif c == curses.KEY_DOWN or c == ord('j'):
                         self.move(1)
                     elif c == ord('f'):
                         if self.current_pad == 'streams':
@@ -330,6 +349,11 @@ class StreamList(object):
                         self.prompt_new_stream()
                     elif c == ord('d'):
                         self.delete_stream()
+                    elif c == ord('o'):
+                        self.show_offline_streams ^= True
+                        self.refilter_streams()
+                    elif c == ord('O'):
+                        self.check_online_streams()
                     elif c == ord('h') or c == ord('?'):
                         self.show_help()
 
@@ -340,7 +364,6 @@ class StreamList(object):
 
         """
         max_y, max_x = self.getheightwidth()
-        #raise ValueError('{} {}'.format(max_y, max_x))
         self.pad_x = 0
         self.max_y, self.max_x = (max_y-1, max_x-1)
         self.pad_w = max_x-1*self.pad_x
@@ -367,8 +390,9 @@ class StreamList(object):
         self.overwrite_line(msg, attr=curses.A_REVERSE)
 
     def init_help(self):
-        help_pad_length = 25    # there should be a neater way to do this
+        help_pad_length = 27    # there should be a neater way to do this
         h = curses.newpad(help_pad_length, self.pad_w)
+        h.keypad(1)
 
         h.addstr( 0, 0, 'STREAM MANAGEMENT', curses.A_BOLD)
         h.addstr( 2, 0, '  Enter : start stream')
@@ -388,10 +412,12 @@ class StreamList(object):
         h.addstr(18, 0, '  k/down: down one line')
         h.addstr(19, 0, '  f     : filter streams')
         h.addstr(20, 0, '  F     : clear filter')
-        h.addstr(21, 0, '  gg    : go to top')
-        h.addstr(22, 0, '  G     : go to bottom')
-        h.addstr(23, 0, '  h/?   : show this help')
-        h.addstr(24, 0, '  q     : quit')
+        h.addstr(21, 0, '  o     : toggle offline streams')
+        h.addstr(22, 0, '  O     : check for online streams')
+        h.addstr(23, 0, '  gg    : go to top')
+        h.addstr(24, 0, '  G     : go to bottom')
+        h.addstr(25, 0, '  h/?   : show this help')
+        h.addstr(26, 0, '  q     : quit')
 
         self.pads['help'] = h
         self.offsets['help'] = 0
@@ -420,6 +446,7 @@ class StreamList(object):
             self.pads['streams'].clear()
             self.refresh_current_pad()
         pad = curses.newpad(max(1,len(self.filtered_streams)), self.pad_w)
+        pad.keypad(1)
         for s in self.filtered_streams:
             pad.addstr(y, 0, self.format_stream_line(s))
             y+=1
@@ -521,14 +548,14 @@ class StreamList(object):
 
     def format_stream_line(self, stream):
         id = '{} '.format(stream['id']).rjust(ID_FIELD_WIDTH)
-        name = ' {}'.format(stream['name']).ljust(NAME_FIELD_WIDTH)
-        res  = ' {}'.format(stream['res']).ljust(RES_FIELD_WIDTH)
+        name = ' {}'.format(stream['name'][:NAME_FIELD_WIDTH-2]).ljust(NAME_FIELD_WIDTH)
+        res  = ' {}'.format(stream['res'][:RES_FIELD_WIDTH-2]).ljust(RES_FIELD_WIDTH)
         views  = '{} '.format(stream['seen']).rjust(VIEWS_FIELD_WIDTH)
         p = self.q.get_process(stream['id']) != None
         if p:
-            indicator = '>'
+            indicator = '[>>>]'
         else:
-            indicator = ' '
+            indicator = INDICATORS[stream['online']]
         return '{}|{}|{}|{}|  {}'.format(id, name, res, views, indicator)
 
     def redraw_current_line(self):
@@ -545,16 +572,21 @@ class StreamList(object):
         pad.move(row, 0)
         self.refresh_current_pad()
 
+    def set_status(self, status):
+        self.status = status
+        self.redraw_status()
+
     def redraw_status(self):
         self.s.move(self.max_y, 0)
         self.overwrite_line(self.status[:self.max_x], curses.A_NORMAL)
+        self.s.refresh()
 
     def redraw_stream_footer(self):
         if not self.no_stream_shown:
             row = self.pads[self.current_pad].getyx()[0]
             s = self.filtered_streams[row]
             self.set_footer('{}/{} {} {}'.format(row+1, len(self.filtered_streams), s['url'], s['res']))
-            self.s.refresh
+            self.s.refresh()
 
     def check_stopped_streams(self):
         finished = self.q.get_finished()
@@ -568,11 +600,51 @@ class StreamList(object):
                             attr = curses.A_REVERSE
                         else:
                             attr = curses.A_NORMAL
-                        self.pads['streams'].addch(i, PLAYING_FIELD_OFFSET, ' ', attr)
+                        self.pads['streams'].addstr(i, PLAYING_FIELD_OFFSET,
+                                                    INDICATORS[s['online']], attr)
                         self.refresh_current_pad()
                 except:
                     pass
 
+
+    def _check_stream(self, url):
+        try:
+            plugin = self.livestreamer.resolve_url(url)
+            avail_streams = plugin.get_streams()
+            if avail_streams:
+                return 1
+            return 0
+        except:
+            return 3
+
+    def check_online_streams(self):
+        self.set_status(' Checking online streams...')
+
+        manager = Manager()
+        queue   = manager.Queue()
+
+        def check_stream_managed(args):
+            url, queue = args
+            status = self._check_stream(url)
+            queue.put(url)
+            return status
+
+        pool = Pool(CHECK_ONLINE_THREADS)
+        args = [(s['url'], queue) for s in self.streams]
+        statuses = pool.map_async(check_stream_managed, args)
+        n_streams = len(self.streams)
+
+        while not statuses.ready():
+            sleep(0.1)
+            self.set_status(' Checked {0}/{1} streams...'.format(queue.qsize(), n_streams))
+            self.s.refresh()
+
+        statuses = statuses.get()
+        for i, s in enumerate(self.streams):
+            s['online'] = statuses[i]
+
+        self.refilter_streams()
+
     def prompt_input(self, prompt=''):
         self.s.move(self.max_y-1, 0)
         self.s.clrtoeol()
@@ -640,13 +712,18 @@ class StreamList(object):
 
     def filter_streams(self):
         self.filter = self.prompt_input('Filter: ').lower()
+        self.refilter_streams()
+
+    def refilter_streams(self):
         self.filtered_streams = []
         for s in self.streams:
-            if self.filter in s['name'].lower() or self.filter in s['url'].lower():
+            if ((self.show_offline_streams or s['online'] in [1,2])
+                and (self.filter in s['name'].lower() or self.filter in s['url'].lower())):
                 self.filtered_streams.append(s)
         self.filtered_streams.sort(key=lambda s:s['seen'], reverse=True)
         self.no_stream_shown = len(self.filtered_streams) == 0
-        self.status = 'New filter: {0} ({1} matches)'.format(self.filter, len(self.filtered_streams))
+        self.status = ' Filter: {0} ({1} matches, {2} showing offline streams)'.format(
+                self.filter or '<empty>', len(self.filtered_streams), '' if self.show_offline_streams else 'NOT')
         self.init_streams_pad()
         self.refresh_current_pad()
         self.redraw_stream_footer()
@@ -663,7 +740,7 @@ class StreamList(object):
                 last_seen = int(time())
             else:
                 seen = last_seen = 0
-            if len(self.streams) == 0:
+            if not self.streams:
                 id = 1
             else:
                 self.max_id += 1
@@ -684,21 +761,22 @@ class StreamList(object):
             else:
                 actual_res = DEFAULT_RESOLUTION_HARD
 
+            self.set_status(' Checking if new stream is online...')
+            self.s.refresh()
+            online = self._check_stream(url)
+
             new_stream = {
                     'id'        : id,
                     'name'      : name,
                     'seen'      : seen,
                     'last_seen' : last_seen,
                     'res'       : actual_res,
-                    'url'       : url
+                    'url'       : url,
+                    'online'    : online
                 }
             self.streams.append(new_stream)
             self.no_streams = False
-            if self.filter in name.lower() or self.filter in url.lower():
-                self.filtered_streams.append(new_stream)
-            self.no_stream_shown = len(self.filtered_streams) == 0
-            try: self.init_streams_pad()
-            except: pass
+            self.refilter_streams()
             self.sync_store()
 
     def delete_stream(self):
@@ -712,9 +790,9 @@ class StreamList(object):
         self.streams.remove(s)
         pad.deleteln()
         self.sync_store()
-        if len(self.streams) == 0:
+        if not self.streams:
             self.no_streams = True
-        if len(self.filtered_streams) == 0:
+        if not self.filtered_streams:
             self.no_stream_shown = True
         if pad.getyx()[0] == len(self.filtered_streams) and not self.no_stream_shown:
             self.move(-1)
@@ -764,7 +842,7 @@ class StreamList(object):
     def prompt_new_stream(self):
         url = self.prompt_input('New stream URL (empty to cancel): ')
         name = url.split('/')[-1]
-        if len(name) > 0:
+        if name:
             self.add_stream(name, url)
             self.move(len(self.filtered_streams)-1, absolute=True)
             self.show_streams()
@@ -809,11 +887,13 @@ def main():
     rc_filename = args.f
     if os.path.exists(rc_filename):
         try:
-            rc = imp.load_source('rc', rc_filename)
+            rc_module = imp.load_source('rc', rc_filename)
         except Exception as e:
             sys.stderr.write('Failed to read rc file, error was:\n{0}\n'.format(str(e)))
             sys.exit(1)
-    l = StreamList(args.d)
+    else:
+        rc_module = imp.new_module('rc')
+    l = StreamList(args.d, rc_module)
     curses.wrapper(l)
 
 if __name__ == '__main__':
